@@ -36,10 +36,6 @@ static const char * const balance_cmd_group_usage[] = {
 	NULL
 };
 
-static const char balance_cmd_group_info[] =
-	"'btrfs filesystem balance' command is deprecated, please use\n"
-	"'btrfs balance start' command instead.";
-
 static int parse_one_profile(const char *profile, u64 *flags)
 {
 	if (!strcmp(profile, "raid0")) {
@@ -727,6 +723,9 @@ static int cmd_balance_status(int argc, char **argv)
 	return 1;
 }
 
+static const char balance_cmd_group_info[] =
+"balance data accross devices, or change block groups using filters";
+
 const struct cmd_group balance_cmd_group = {
 	balance_cmd_group_usage, balance_cmd_group_info, {
 		{ "start", cmd_balance_start, cmd_balance_start_usage, NULL, 0 },
